@@ -1,22 +1,23 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2020-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2020-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
 import signal
-from gi.repository import Gtk, Gdk, GLib, Pango
+from time import sleep, monotonic
+from gi.repository import Gtk, Gdk, GLib, Pango  # @UnresolvedImport
 
 from xpra import __version__
-from xpra.util import envint
-from xpra.os_util import SIGNAMES, OSX
+from xpra.util import envint, envbool
+from xpra.os_util import SIGNAMES, OSX, WIN32
 from xpra.exit_codes import EXIT_TIMEOUT, EXIT_CONNECTION_LOST
 from xpra.common import SPLASH_EXIT_DELAY
 from xpra.gtk_common.gtk_util import add_close_accel, get_icon_pixbuf
 from xpra.gtk_common.gobject_compat import install_signal_handlers
 from xpra.client.gtk_base.css_overrides import inject_css_overrides
-from xpra.platform.gui import force_focus
+from xpra.platform.gui import force_focus, set_window_progress
 from xpra.log import Logger
 
 log = Logger("client", "util")
@@ -25,6 +26,8 @@ inject_css_overrides()
 
 TIMEOUT = envint("XPRA_SPLASH_TIMEOUT", 60)
 LINES = envint("XPRA_SPLASH_LINES", 4)
+READ_SLEEP = envint("XPRA_SPLASH_READ_SLEEP", 0)
+FOCUS_EXIT = envbool("XPRA_SPLASH_FOCUS_EXIT", True)
 
 #PULSE_CHARS = "▁▂▃▄▅▆▇█▇▆▅▄▃▁"
 PULSE_CHARS = "◐◓◑◒"
@@ -94,6 +97,9 @@ class SplashScreen(Gtk.Window):
             signal.signal(SIGPIPE, self.handle_signal)
         self.opacity = 100
         self.pct = 0
+        self.start_time = monotonic()
+        self.had_top_level_focus = False
+        self.connect("notify::has-toplevel-focus", self._focus_change)
 
 
     def run(self):
@@ -110,9 +116,23 @@ class SplashScreen(Gtk.Window):
                 import ctypes  #pylint: disable=import-outside-toplevel
                 ctypes.string_at(0)
             GLib.timeout_add(scrash, crash)
+        self.start_time = monotonic()
         Gtk.main()
         return self.exit_code or 0
 
+    def _focus_change(self, *args):
+        if WIN32 and monotonic()-self.start_time<1:
+            #ignore initial focus events on win32
+            return
+        has = self.has_toplevel_focus()
+        had = self.had_top_level_focus
+        log(f"_focus_change{args} had={had}, has={has}")
+        if had and not has:
+            if FOCUS_EXIT:
+                self.exit()
+        elif has:
+            self.had_top_level_focus = True
+
     def timeout(self):
         log("timeout()")
         self.timeout_timer = None
@@ -120,6 +140,7 @@ class SplashScreen(Gtk.Window):
         self.show_progress_value(100)
         self.progress_bar.set_text("timeout")
         self.progress_bar.set_show_text(True)
+        self.exit()
 
     def cancel_timeout_timer(self):
         tt = self.timeout_timer
@@ -146,8 +167,9 @@ class SplashScreen(Gtk.Window):
             if not line:
                 self.exit_code = EXIT_CONNECTION_LOST
                 self.exit()
-            else:
-                GLib.idle_add(self.handle_stdin_line, line)
+                break
+            GLib.idle_add(self.handle_stdin_line, line)
+            sleep(READ_SLEEP)
 
     def handle_stdin_line(self, line):
         parts = line.rstrip("\n\r").split(":", 1)
@@ -196,6 +218,7 @@ class SplashScreen(Gtk.Window):
             self.progress_timer = GLib.timeout_add(40, self.increase_fraction, pct)
             self.opacity = min(100, max(50, 130-pct))
             self.set_opacity(self.opacity/100.0)
+        set_window_progress(self, pct)
 
     def cancel_exit_timer(self):
         et = self.exit_timer
