@@ -1,6 +1,6 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2022 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -15,14 +15,14 @@ import math
 from collections import deque
 from time import sleep, time, monotonic
 from queue import Queue
-from gi.repository import GLib
+from gi.repository import GLib  # @UnresolvedImport
 
 from xpra.platform.gui import (
     get_window_min_size, get_window_max_size,
     get_double_click_time, get_double_click_distance, get_native_system_tray_classes,
     )
-from xpra.platform.features import SYSTEM_TRAY_SUPPORTED
-from xpra.platform.paths import get_icon_filename
+from xpra.common import WINDOW_NOT_FOUND, WINDOW_DECODE_SKIPPED, WINDOW_DECODE_ERROR
+from xpra.platform.paths import get_icon_filename, get_resources_dir, get_python_exec_command
 from xpra.scripts.config import FALSE_OPTIONS
 from xpra.make_thread import start_thread
 from xpra.os_util import (
@@ -49,10 +49,11 @@ metalog = Logger("metadata")
 traylog = Logger("client", "tray")
 
 MOUSE_SHOW = envbool("XPRA_MOUSE_SHOW", True)
+SMOOTH_SCROLL = envbool("XPRA_SMOOTH_SCROLL", True)
 
 PAINT_FAULT_RATE = envint("XPRA_PAINT_FAULT_INJECTION_RATE")
 PAINT_FAULT_TELL = envbool("XPRA_PAINT_FAULT_INJECTION_TELL", True)
-PAINT_DELAY = envint("XPRA_PAINT_DELAY", 0)
+PAINT_DELAY = envint("XPRA_PAINT_DELAY", -1)
 
 WM_CLASS_CLOSEEXIT = os.environ.get("XPRA_WM_CLASS_CLOSEEXIT", "Xephyr").split(",")
 TITLE_CLOSEEXIT = os.environ.get("XPRA_TITLE_CLOSEEXIT", "Xnest").split(",")
@@ -73,12 +74,24 @@ ICON_OVERLAY = envint("XPRA_ICON_OVERLAY", 50)
 ICON_SHRINKAGE = envint("XPRA_ICON_SHRINKAGE", 75)
 SAVE_WINDOW_ICONS = envbool("XPRA_SAVE_WINDOW_ICONS", False)
 SAVE_CURSORS = envbool("XPRA_SAVE_CURSORS", False)
-SIGNAL_WATCHER = envbool("XPRA_SIGNAL_WATCHER", True)
+SIGNAL_WATCHER = envbool("XPRA_SIGNAL_WATCHER", POSIX and not OSX)
+SIGNAL_WATCHER_COMMAND = os.environ.get("XPRA_SIGNAL_WATCHER_COMMAND", "xpra_signal_listener")
+if SIGNAL_WATCHER:
+    SIGNAL_WATCHER = False
+    for prefix in ("/usr", get_resources_dir()):
+        cmd = prefix+"/libexec/xpra/"+SIGNAL_WATCHER_COMMAND
+        if os.path.exists(cmd):
+            SIGNAL_WATCHER_COMMAND = cmd
+            SIGNAL_WATCHER = True
+    if not SIGNAL_WATCHER:
+        log.warn("Warning: %r not found", SIGNAL_WATCHER_COMMAND)
 
 FAKE_SUSPEND_RESUME = envint("XPRA_FAKE_SUSPEND_RESUME", 0)
 MOUSE_SCROLL_SQRT_SCALE = envbool("XPRA_MOUSE_SCROLL_SQRT_SCALE", OSX)
 MOUSE_SCROLL_MULTIPLIER = envint("XPRA_MOUSE_SCROLL_MULTIPLIER", 100)
 
+PRE_MAP = envbool("XPRA_PRE_MAP_WINDOWS", True)
+SHOW_DELAY = envint("XPRA_SHOW_DELAY", -1)
 
 DRAW_TYPES = {bytes : "bytes", str : "bytes", tuple : "arrays", list : "arrays"}
 
@@ -139,6 +152,7 @@ class WindowClient(StubClientMixin):
         self._pid_to_signalwatcher = {}
         self._signalwatcher_to_wids = {}
 
+        self.wheel_smooth = SMOOTH_SCROLL
         self.wheel_map = {}
         self.wheel_deltax = 0
         self.wheel_deltay = 0
@@ -147,11 +161,12 @@ class WindowClient(StubClientMixin):
         self.lost_focus_timer = None
         self._focused = None
         self._window_with_grab = None
+        self.pointer_grabbed = None
         self._suspended_at = 0
         self._button_state = {}
 
     def init(self, opts):
-        if opts.system_tray and SYSTEM_TRAY_SUPPORTED:
+        if opts.system_tray:
             try:
                 from xpra.client import client_tray
                 assert client_tray
@@ -195,19 +210,22 @@ class WindowClient(StubClientMixin):
             self.parse_border()
 
         #mouse wheel:
-        mw = (opts.mousewheel or "").lower().replace("-", "")
-        if mw not in FALSE_OPTIONS:
+        mw = (opts.mousewheel or "").lower().replace("-", "").split(",")
+        if "coarse" in mw:
+            mw.remove("coarse")
+            self.wheel_smooth = False
+        if not any(x in FALSE_OPTIONS for x in mw):
             UP = 4
             LEFT = 6
             Z1 = 8
+            invertall = len(mw)==1 and mw[0] in ("invert", "invertall")
             for i in range(20):
                 btn = 4+i*2
                 invert = (
-                    mw=="invert" or
-                    mw=="invertall" or
-                    (btn==UP and mw=="inverty") or
-                    (btn==LEFT and mw=="invertx") or
-                    (btn==Z1 and mw=="invertz")
+                    invertall or
+                    (btn==UP and "inverty" in mw) or
+                    (btn==LEFT and "invertx" in mw) or
+                    (btn==Z1 and "invertz" in mw)
                     )
                 if not invert:
                     self.wheel_map[btn] = btn
@@ -215,7 +233,7 @@ class WindowClient(StubClientMixin):
                 else:
                     self.wheel_map[btn+1] = btn
                     self.wheel_map[btn] = btn+1
-        mouselog("wheel_map(%s)=%s", mw, self.wheel_map)
+        mouselog("wheel_map(%s)=%s, wheel_smooth=%s", mw, self.wheel_map, self.wheel_smooth)
 
         if 0<ICON_OVERLAY<=100:
             icon_filename = opts.tray_icon
@@ -297,6 +315,7 @@ class WindowClient(StubClientMixin):
             },
             "focused"       : self._focused or 0,
             "grabbed"       : self._window_with_grab or 0,
+            "pointer-grab"  : self.pointer_grabbed or 0,
             "buttons"       : self._button_state,
         }
         for wid, window in tuple(self._id_to_window.items()):
@@ -337,6 +356,7 @@ class WindowClient(StubClientMixin):
             "min-size"                  : self.min_window_size,
             "max-size"                  : self.max_window_size,
             "restack"                   : True,
+            "pre-map"                   : PRE_MAP,
             }
 
 
@@ -471,7 +491,7 @@ class WindowClient(StubClientMixin):
             new_cursor = packet[1]
         else:
             if len(packet)<9:
-                raise Exception("invalid cursor packet: %s items" % len(packet))
+                raise Exception(f"invalid cursor packet: {len(packet)} items")
             #trim packet-type:
             new_cursor = packet[1:]
             encoding = u(new_cursor[0])
@@ -480,7 +500,7 @@ class WindowClient(StubClientMixin):
                 pixels = new_cursor[8]
                 if SAVE_CURSORS:
                     serial = new_cursor[7]
-                    with open("raw-cursor-%#x.png" % serial, 'wb') as f:
+                    with open(f"raw-cursor-{serial:x}.png", "wb") as f:
                         f.write(pixels)
                 from xpra.codecs.pillow.decoder import open_only  #pylint: disable=import-outside-toplevel
                 img = open_only(pixels, ("png",))
@@ -488,7 +508,7 @@ class WindowClient(StubClientMixin):
                 cursorlog("used PIL to convert png cursor to raw")
                 new_cursor[0] = "raw"
             elif encoding!="raw":
-                cursorlog.warn("Warning: invalid cursor encoding: %s", encoding)
+                cursorlog.warn(f"Warning: invalid cursor encoding: {encoding}")
                 return
         self.set_windows_cursor(self._id_to_window.values(), new_cursor)
 
@@ -658,7 +678,7 @@ class WindowClient(StubClientMixin):
     def _window_icon_image(self, wid, width, height, coding, data):
         #convert the data into a pillow image,
         #adding the icon overlay (if enabled)
-        from PIL import Image
+        from PIL import Image  # @UnresolvedImport
         coding = bytestostr(coding)
         iconlog("%s.update_icon(%s, %s, %s, %s bytes) ICON_SHRINKAGE=%s, ICON_OVERLAY=%s",
                 self, width, height, coding, len(data), ICON_SHRINKAGE, ICON_OVERLAY)
@@ -668,11 +688,7 @@ class WindowClient(StubClientMixin):
             rowstride = width*4
             img = Image.frombytes("RGBA", (width,height), memoryview_to_bytes(data), "raw", "BGRA", rowstride, 1)
             has_alpha = True
-        elif coding in ("BGRA", "premult_argb32"):
-            if coding == "premult_argb32":
-                #we usually cannot do in-place and this is not performance critical
-                from xpra.codecs.argb.argb import unpremultiply_argb    #@UnresolvedImport
-                data = unpremultiply_argb(data)
+        elif coding in ("BGRA", ):
             rowstride = width*4
             img = Image.frombytes("RGBA", (width,height), memoryview_to_bytes(data), "raw", "BGRA", rowstride, 1)
             has_alpha = True
@@ -733,6 +749,17 @@ class WindowClient(StubClientMixin):
         if w<1 or h<1:
             log.error("Error: window %i dimensions %ix%i are invalid", wid, w, h)
             w, h = 1, 1
+        rel_pos = metadata.get("relative-position")
+        parent = metadata.get("parent")
+        geomlog("relative-position=%s (parent=%s)", rel_pos, parent)
+        if parent and rel_pos:
+            pwin = self._id_to_window.get(parent)
+            if pwin:
+                #apply scaling to relative position:
+                p_pos = pwin.sp(*rel_pos)
+                x = pwin._pos[0] + p_pos[0]
+                y = pwin._pos[1] + p_pos[1]
+                geomlog("relative position(%s)=%s", rel_pos, (x, y))
         #scaled dimensions of window:
         wx = self.sx(x)
         wy = self.sy(y)
@@ -745,7 +772,7 @@ class WindowClient(StubClientMixin):
             client_properties = packet[7]
         geomlog("process_new_common: wid=%i, OR=%s, geometry(%s)=%s / %s",
                 wid, override_redirect, packet[2:6], (wx, wy, ww, wh), (bw, bh))
-        self.make_new_window(wid, wx, wy, ww, wh, bw, bh, metadata, override_redirect, client_properties)
+        return self.make_new_window(wid, wx, wy, ww, wh, bw, bh, metadata, override_redirect, client_properties)
 
     def make_new_window(self, wid, wx, wy, ww, wh, bw, bh, metadata, override_redirect, client_properties):
         client_window_classes = self.get_client_window_classes(ww, wh, metadata, override_redirect)
@@ -788,12 +815,18 @@ class WindowClient(StubClientMixin):
         log("make_new_window(..) window(%i)=%s", wid, window)
         self._id_to_window[wid] = window
         self._window_to_id[window] = wid
+        if SHOW_DELAY>=0:
+            self.timeout_add(SHOW_DELAY, self.show_window, wid, window, metadata, override_redirect)
+        else:
+            self.show_window(wid, window, metadata, override_redirect)
+        return window
+
+    def show_window(self, wid, window, metadata, override_redirect):
         window.show_all()
         if override_redirect:
             if self.should_force_grab(metadata):
                 grablog.warn("forcing grab for OR window %i, matches %s", wid, OR_FORCE_GRAB)
                 self.window_grab(wid, window)
-        return window
 
     def should_force_grab(self, metadata):
         if not OR_FORCE_GRAB:
@@ -815,16 +848,14 @@ class WindowClient(StubClientMixin):
     ######################################################################
     # listen for process signals using a watcher process:
     def assign_signal_watcher_pid(self, wid, pid):
-        if not SIGNAL_WATCHER:
-            return 0
-        if not POSIX or OSX or not pid:
+        if not SIGNAL_WATCHER or not pid:
             return 0
         proc = self._pid_to_signalwatcher.get(pid)
         if proc is None or proc.poll():
             from xpra.child_reaper import getChildReaper
             from subprocess import Popen, PIPE, STDOUT
             try:
-                proc = Popen(["xpra_signal_listener"],
+                proc = Popen(get_python_exec_command()+[SIGNAL_WATCHER_COMMAND],
                              stdin=PIPE, stdout=PIPE, stderr=STDOUT,
                              start_new_session=True)
             except OSError as e:
@@ -1011,7 +1042,7 @@ class WindowClient(StubClientMixin):
 
 
     def _process_new_window(self, packet):
-        self._process_new_common(packet, False)
+        return self._process_new_common(packet, False)
 
     def _process_new_override_redirect(self, packet):
         if self.modal_windows:
@@ -1024,7 +1055,7 @@ class WindowClient(StubClientMixin):
                 if window.get_modal():
                     metalog("temporarily removing modal flag from %s", wid)
                     window.set_modal(False)
-        self._process_new_common(packet, True)
+        return self._process_new_common(packet, True)
 
 
     def _process_initiate_moveresize(self, packet):
@@ -1179,6 +1210,8 @@ class WindowClient(StubClientMixin):
             log("destroying window %s which has grab, ungrabbing!", wid)
             self.window_ungrab()
             self._window_with_grab = None
+        if self.pointer_grabbed==wid:
+            self.pointer_grabbed = None
         #deal with signal watchers:
         log("looking for window %i in %s", wid, self._signalwatcher_to_wids)
         for signalwatcher, wids in tuple(self._signalwatcher_to_wids.items()):
@@ -1256,9 +1289,10 @@ class WindowClient(StubClientMixin):
         self.send("focus", wid, self.get_current_modifiers())
 
     def update_focus(self, wid, gotit):
-        focuslog("update_focus(%s, %s) focused=%s, grabbed=%s", wid, gotit, self._focused, self._window_with_grab)
+        focused = self._focused
+        focuslog(f"update_focus({wid}, {gotit}) focused={focused}, grabbed={self._window_with_grab}")
         if gotit:
-            if self._focused is not wid:
+            if focused is not wid:
                 self.send_focus(wid)
                 self._focused = wid
             self.cancel_lost_focus_timer()
@@ -1269,17 +1303,18 @@ class WindowClient(StubClientMixin):
                 if wwgrab:
                     self.do_force_ungrab(wwgrab)
                 self._window_with_grab = None
-            if wid and self._focused and self._focused!=wid:
+            if wid and focused and focused!=wid:
                 #if this window lost focus, it must have had it!
                 #(catch up - makes things like OR windows work:
                 # their parent receives the focus-out event)
-                focuslog("window %s lost a focus it did not have!? (simulating focus before losing it)", wid)
+                focuslog(f"window {wid} lost a focus it did not have!? (simulating focus before losing it)")
                 self.send_focus(wid)
-            if self._focused and not self.lost_focus_timer:
+            if focused and not self.lost_focus_timer:
                 #send the lost-focus via a timer and re-check it
                 #(this allows a new window to gain focus without having to do a reset_focus)
                 self.lost_focus_timer = self.timeout_add(20, self.send_lost_focus)
                 self._focused = None
+        return focused!=self._focused
 
     def send_lost_focus(self):
         focuslog("send_lost_focus() focused=%s", self._focused)
@@ -1392,7 +1427,7 @@ class WindowClient(StubClientMixin):
     ######################################################################
     # painting windows:
     def _process_draw(self, packet):
-        if PAINT_DELAY>0:
+        if PAINT_DELAY>=0:
             self.timeout_add(PAINT_DELAY, self._draw_queue.put, packet)
         else:
             self._draw_queue.put(packet)
@@ -1443,7 +1478,7 @@ class WindowClient(StubClientMixin):
                     data_start.value = offset+length
                     #clear the mmap area via idle_add so any pending draw requests
                     #will get a chance to run first (preserving the order)
-                self.send_damage_sequence(wid, packet_sequence, width, height, -1)
+                self.send_damage_sequence(wid, packet_sequence, width, height, WINDOW_NOT_FOUND, "window not found")
             self.idle_add(draw_cleanup)
             return
         #rename old encoding aliases early:
@@ -1464,12 +1499,12 @@ class WindowClient(StubClientMixin):
                 paintlog("record_decode_time(%s, %s) wid=%s, %s: %sx%s, %s",
                          success, message, wid, coding, width, height, dms)
             elif success==0:
-                decode_time = -1
+                decode_time = WINDOW_DECODE_ERROR
                 paintlog("record_decode_time(%s, %s) decoding error on wid=%s, %s: %sx%s",
                          success, message, wid, coding, width, height)
             else:
                 assert success<0
-                decode_time = 0
+                decode_time = WINDOW_DECODE_SKIPPED
                 paintlog("record_decode_time(%s, %s) decoding or painting skipped on wid=%s, %s: %sx%s",
                          success, message, wid, coding, width, height)
             self.send_damage_sequence(wid, packet_sequence, width, height, decode_time, repr_ellipsized(message, 512))
