@@ -1,29 +1,39 @@
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
+import os
 import sys
 
 from xpra.client.client_base import XpraClientBase
 from xpra.client.keyboard_helper import KeyboardHelper
 from xpra.platform import set_name
 from xpra.platform.gui import ready as gui_ready, get_wm_name, get_session_type, ClientExtras
+from xpra.common import FULL_INFO
 from xpra.version_util import full_version_str
 from xpra.net import compression, packet_encoding
+from xpra.net.net_util import get_info as get_net_info
 from xpra.child_reaper import reaper_cleanup
-from xpra.os_util import platform_name, bytestostr, strtobytes, BITS, POSIX, is_Wayland
+from xpra.platform.info import get_sys_info
+from xpra.os_util import (
+    platform_name, bytestostr,
+    BITS, POSIX, WIN32, OSX, is_Wayland,
+    get_frame_info, get_info_env, get_sysconfig_info,
+    )
 from xpra.util import (
     std, envbool, envint, typedict, updict, repr_ellipsized, ellipsizer, log_screen_sizes, engs, csv,
     merge_dicts,
     XPRA_AUDIO_NOTIFICATION_ID, XPRA_DISCONNECT_NOTIFICATION_ID,
+    AUTHENTICATION_FAILED,
     )
-from xpra.exit_codes import EXIT_CONNECTION_FAILED, EXIT_OK, EXIT_CONNECTION_LOST
+from xpra.scripts.config import parse_bool
+from xpra.exit_codes import EXIT_CONNECTION_FAILED, EXIT_CONNECTION_LOST, EXIT_AUTHENTICATION_FAILED
 from xpra.version_util import get_version_info_full, get_platform_info
 from xpra.client import mixin_features
-from xpra.log import Logger
+from xpra.log import Logger, get_info as get_log_info
 
 
 CLIENT_BASES = [XpraClientBase]
@@ -76,6 +86,7 @@ log("UIXpraClient%s: %s", ClientBaseClass, CLIENT_BASES)
 
 NOTIFICATION_EXIT_DELAY = envint("XPRA_NOTIFICATION_EXIT_DELAY", 2)
 MOUSE_DELAY_AUTO = envbool("XPRA_MOUSE_DELAY_AUTO", True)
+SYSCONFIG = envbool("XPRA_SYSCONFIG", FULL_INFO>0)
 
 
 """
@@ -93,7 +104,7 @@ class UIXpraClient(ClientBaseClass):
             __signals__ += c.__signals__
 
     def __init__(self):
-        log.info("Xpra %s client version %s %i-bit", self.client_toolkit(), full_version_str(), BITS)
+        log.info(f"Xpra {self.client_toolkit()} client version {full_version_str()} {BITS}-bit")
         #mmap_enabled belongs in the MmapClient mixin,
         #but it is used outside it, so make sure we define it:
         self.mmap_enabled = False
@@ -105,12 +116,12 @@ class UIXpraClient(ClientBaseClass):
         try:
             pinfo = get_platform_info()
             osinfo = "%s" % platform_name(sys.platform, pinfo.get("linux_distribution") or pinfo.get("sysrelease", ""))
-            log.info(" running on %s", osinfo)
+            log.info(f" running on {osinfo}")
         except Exception:
             log("platform name error:", exc_info=True)
         wm = get_wm_name()      #pylint: disable=assignment-from-none
         if wm:
-            log.info(" window manager is '%s'", wm)
+            log.info(f" window manager is {wm!r}")
 
         self._ui_events = 0
         self.title = ""
@@ -126,8 +137,10 @@ class UIXpraClient(ClientBaseClass):
         self.xsettings_enabled = False
         self.server_start_new_commands = False
         self.server_xdg_menu = None
-        self.start_new_commands  = []
-        self.start_child_new_commands  = []
+        self.start_new_commands = False
+        self.request_start = []
+        self.request_start_child = []
+        self.headerbar = None
 
         #in WindowClient - should it be?
         #self.server_is_desktop = False
@@ -148,7 +161,6 @@ class UIXpraClient(ClientBaseClass):
         self.keyboard_helper = None
         self.keyboard_grabbed = False
         self.keyboard_sync = False
-        self.pointer_grabbed = False
         self.kh_warning = False
         self.menu_helper = None
 
@@ -159,16 +171,18 @@ class UIXpraClient(ClientBaseClass):
 
     def init(self, opts):
         """ initialize variables from configuration """
+        self.init_aliases()
         for c in CLIENT_BASES:
-            log("init: %s", c)
+            log(f"init: {c}")
             c.init(self, opts)
 
         self.title = opts.title
         self.session_name = bytestostr(opts.session_name)
-        self.xsettings_enabled = opts.xsettings
+        self.xsettings_enabled = not (OSX or WIN32) and parse_bool("xsettings", opts.xsettings, True)
         self.readonly = opts.readonly
         self.client_supports_sharing = opts.sharing is True
         self.client_lock = opts.lock is True
+        self.headerbar = opts.headerbar
 
 
     def init_ui(self, opts):
@@ -192,7 +206,7 @@ class UIXpraClient(ClientBaseClass):
                                                                   opts.keyboard_raw, *overrides)
             except ImportError as e:
                 keylog("error instantiating %s", self.keyboard_helper_class, exc_info=True)
-                keylog.warn("Warning: no keyboard support, %s", e)
+                keylog.warn(f"Warning: no keyboard support, {e}")
 
         if mixin_features.windows:
             self.init_opengl(opts.opengl)
@@ -200,26 +214,30 @@ class UIXpraClient(ClientBaseClass):
         if ClientExtras is not None:
             self.client_extras = ClientExtras(self, opts)   #pylint: disable=not-callable
 
-        if opts.start or opts.start_child:
+        self.start_new_commands = parse_bool("start-new-commands", opts.start_new_commands, True)
+        if self.start_new_commands and (opts.start or opts.start_child):
             from xpra.scripts.main import strip_defaults_start_child
             from xpra.scripts.config import make_defaults_struct
             defaults = make_defaults_struct()
-            self.start_new_commands  = strip_defaults_start_child(opts.start, defaults.start)   #pylint: disable=no-member
-            self.start_child_new_commands  = strip_defaults_start_child(opts.start_child, defaults.start_child) #pylint: disable=no-member
+            self.request_start = strip_defaults_start_child(opts.start, defaults.start)   #pylint: disable=no-member
+            self.request_start_child = strip_defaults_start_child(opts.start_child, defaults.start_child) #pylint: disable=no-member
 
         if MOUSE_DELAY_AUTO:
             try:
-                from xpra.platform.gui import get_vrefresh
-                v = get_vrefresh()
-                if v<=0:
-                    #some platforms don't detect the vrefresh correctly
-                    #(ie: macos in virtualbox?), so use a sane default:
-                    v = 60
-                self._mouse_position_delay = 1000//v//2
-                log("mouse delay: %s", self._mouse_position_delay)
+                #some platforms don't detect the vrefresh correctly
+                #(ie: macos in virtualbox?), so use a sane default minimum
+                #discount by 5ms to ensure we have time to hit the target
+                v = max(60, self.get_vrefresh())
+                self._mouse_position_delay = max(5, 1000//v//2 - 5)
+                log(f"mouse position delay: {self._mouse_position_delay}")
             except Exception:
                 log("failed to calculate automatic delay", exc_info=True)
 
+    def get_vrefresh(self):
+        #this method is overriden in the GTK client
+        from xpra.platform.gui import get_vrefresh  #pylint: disable=import-outside-toplevel
+        return get_vrefresh()
+
 
     def run(self):
         if self.client_extras:
@@ -228,7 +246,7 @@ class UIXpraClient(ClientBaseClass):
             c.run(self)
 
 
-    def quit(self, _exit_code=0):
+    def quit(self, exit_code=0):
         raise NotImplementedError()
 
     def cleanup(self):
@@ -238,11 +256,11 @@ class UIXpraClient(ClientBaseClass):
         for x in (self.keyboard_helper, self.tray, self.menu_helper, self.client_extras):
             if x is None:
                 continue
-            log("UIXpraClient.cleanup() calling %s.cleanup()", type(x))
+            log(f"UIXpraClient.cleanup() calling {type(x)}.cleanup()")
             try:
                 x.cleanup()
             except Exception:
-                log.error("error on %s cleanup", type(x), exc_info=True)
+                log.error(f"error on {type(x)} cleanup", exc_info=True)
         #the protocol has been closed, it is now safe to close all the windows:
         #(cleaner and needed when we run embedded in the client launcher)
         reaper_cleanup()
@@ -258,23 +276,34 @@ class UIXpraClient(ClientBaseClass):
 
     def get_info(self):
         info = {}
+        if FULL_INFO>0:
+            info.update({
+                "pid"       : os.getpid(),
+                "sys"       : get_sys_info(),
+                "network"   : get_net_info(),
+                "logging"   : get_log_info(),
+                "threads"   : get_frame_info(),
+                "env"       : get_info_env(),
+                })
+        if SYSCONFIG:
+            info["sysconfig"] = get_sysconfig_info()
         for c in CLIENT_BASES:
             try:
                 i = c.get_info(self)
                 info = merge_dicts(info, i)
             except Exception:
-                log.error("Error collection information from %s", c, exc_info=True)
+                log.error(f"Error collection information from {c}", exc_info=True)
         return info
 
 
     def show_about(self, *_args):
-        log.warn("show_about() is not implemented in %s", self)
+        log.warn(f"show_about() is not implemented in {self!r}")
 
     def show_session_info(self, *_args):
-        log.warn("show_session_info() is not implemented in %s", self)
+        log.warn(f"show_session_info() is not implemented in {self!r}")
 
     def show_bug_report(self, *_args):
-        log.warn("show_bug_report() is not implemented in %s", self)
+        log.warn(f"show_bug_report() is not implemented in {self!r}")
 
 
     def init_opengl(self, _enable_opengl):
@@ -297,23 +326,24 @@ class UIXpraClient(ClientBaseClass):
 
 
     def send_start_new_commands(self):
-        log("send_start_new_commands() start_new_commands=%s, start_child_new_commands=%s",
-            self.start_new_commands, self.start_child_new_commands)
+        log(f"send_start_new_commands() request_start={self.request_start}, request_start_child={self.request_start_child}")
         import shlex
-        for cmd in self.start_new_commands:
+        for cmd in self.request_start:
             cmd_parts = shlex.split(cmd)
             self.send_start_command(cmd_parts[0], cmd, True)
-        for cmd in self.start_child_new_commands:
+        for cmd in self.request_start_child:
             cmd_parts = shlex.split(cmd)
             self.send_start_command(cmd_parts[0], cmd, False)
 
     def send_start_command(self, name, command, ignore, sharing=True):
-        log("send_start_command(%s, %s, %s, %s)", name, command, ignore, sharing)
+        log("send_start_command%s", (name, command, ignore, sharing))
         assert name is not None and command is not None and ignore is not None
         self.send("start-command", name, command, ignore, sharing)
 
     def get_version_info(self) -> dict:
-        return get_version_info_full()
+        if FULL_INFO>1:
+            return get_version_info_full()
+        return super().get_version_info()
 
 
     ######################################################################
@@ -322,11 +352,17 @@ class UIXpraClient(ClientBaseClass):
     def server_disconnect_warning(self, reason, *info):
         if self.exit_code is None:
             body = "\n".join(info)
-            if self.connection_established:
+            if AUTHENTICATION_FAILED in info and not self.connection_established:
+                title = "Authentication failed"
+                self.exit_code = EXIT_AUTHENTICATION_FAILED
+            if self.connection_established and self.completed_startup:
                 title = "Xpra Session Disconnected: %s" % reason
                 self.exit_code = EXIT_CONNECTION_LOST
+            elif self.connection_established:
+                title = "Connection failed during startup: %s" % reason
+                self.exit_code = EXIT_CONNECTION_FAILED
             else:
-                title = "Connection Failed: %s" % reason
+                title = "Connection failed: %s" % reason
                 self.exit_code = EXIT_CONNECTION_FAILED
             self.may_notify(XPRA_DISCONNECT_NOTIFICATION_ID,
                             title, body, icon_name="disconnected")
@@ -338,9 +374,10 @@ class UIXpraClient(ClientBaseClass):
     def server_disconnect(self, reason, *info):
         body = "\n".join(info)
         self.may_notify(XPRA_DISCONNECT_NOTIFICATION_ID,
-                        "Xpra Session Disconnected: %s" % reason, body, icon_name="disconnected")
-        self.exit_code = EXIT_OK
+                        f"Xpra Session Disconnected: {reason}", body, icon_name="disconnected")
         delay = NOTIFICATION_EXIT_DELAY*mixin_features.notifications
+        if self.exit_code is None:
+            self.exit_code = self.server_disconnect_exit_code(reason, *info)
         self.timeout_add(delay*1000, XpraClientBase.server_disconnect, self, reason, *info)
         self.cleanup()
 
@@ -349,35 +386,40 @@ class UIXpraClient(ClientBaseClass):
     # hello:
     def make_hello(self):
         caps = XpraClientBase.make_hello(self)
-        caps["session-type"] = get_session_type()
         #don't try to find the server uuid if this platform cannot run servers..
         #(doing so causes lockups on win32 and startup errors on osx)
         if POSIX and not is_Wayland():
             #we may be running inside another server!
             try:
-                from xpra.server.server_uuid import get_uuid
-                caps["server_uuid"] = get_uuid() or ""
+                from xpra.server.server_uuid import get_uuid, get_mode  #pylint: disable=import-outside-toplevel
+                if get_mode()!="shadow":
+                    uuid = get_uuid()
+                    if uuid:
+                        caps["server_uuid"] = uuid
             except ImportError:
                 pass
         for x in (#generic feature flags:
             "wants_events", "setting-change",
-            "xdg-menu-update",
+            "xdg-menu-update", "mouse",
             ):
             caps[x] = True
+        caps.setdefault("wants", []).append("events")
         caps.update({
             #generic server flags:
             "share"                     : self.client_supports_sharing,
             "lock"                      : self.client_lock,
+            "xdg-menu"                  : self.start_new_commands,
             })
-        caps.update({"mouse" : True})
         caps.update(self.get_keyboard_caps())
         for c in CLIENT_BASES:
             caps.update(c.get_caps(self))
         def u(prefix, c):
             updict(caps, prefix, c, flatten_dicts=False)
         u("control_commands",   self.get_control_commands_caps())
-        u("platform",           get_platform_info())
-        u("opengl",             self.opengl_props)
+        if FULL_INFO>0:
+            u("platform",           get_platform_info())
+            u("opengl",             self.opengl_props)
+            caps["session-type"] = get_session_type()
         return caps
 
 
@@ -402,9 +444,9 @@ class UIXpraClient(ClientBaseClass):
     def parse_server_capabilities(self, c : typedict) -> bool:
         for cb in CLIENT_BASES:
             if not cb.parse_server_capabilities(self, c):
-                log.info("failed to parse server capabilities in %s", cb)
+                log.info(f"failed to parse server capabilities in {cb}")
                 return False
-        self.server_session_name = strtobytes(c.rawget("session_name", b"")).decode("utf-8")
+        self.server_session_name = c.uget("session_name")
         set_name("Xpra", self.session_name or self.server_session_name or "Xpra")
         self.server_platform = c.strget("platform")
         self.server_sharing = c.boolget("sharing")
@@ -416,7 +458,7 @@ class UIXpraClient(ClientBaseClass):
         self.server_start_new_commands = c.boolget("start-new-commands")
         if self.server_start_new_commands:
             self.server_xdg_menu = c.dictget("xdg-menu", None)
-        if self.start_new_commands or self.start_child_new_commands:
+        if self.request_start or self.request_start_child:
             if self.server_start_new_commands:
                 self.after_handshake(self.send_start_new_commands)
             else:
@@ -440,20 +482,19 @@ class UIXpraClient(ClientBaseClass):
         if self._remote_revision:
             r += "-r%s" % self._remote_revision
         mode = c.strget("server.mode", "server")
-        bits = c.intget("python.bits", 32)
-        log.info("Xpra %s server version %s %i-bit", mode, std(r), bits)
-        if i:
-            log.info(" running on %s", std(i))
+        bits = c.intget("python.bits", 0)
+        bitsstr = "" if bits==0 else f" {bits}-bit"
+        log.info(f"Xpra {mode} server version {std(r)}{bitsstr}")
+        if i and i!="unknown":
+            log.info(f" running on {std(i)}")
         if c.boolget("desktop") or c.boolget("shadow"):
             v = c.intpair("actual_desktop_size")
             if v:
                 w, h = v
                 ss = c.tupleget("screen_sizes")
+                log.info(f" remote desktop size is {w}x{h}")
                 if ss:
-                    log.info(" remote desktop size is %sx%s with %s screen%s:", w, h, len(ss), engs(ss))
                     log_screen_sizes(w, h, ss)
-                else:
-                    log.info(" remote desktop size is %sx%s", w, h)
         if c.boolget("proxy"):
             proxy_hostname = c.strget("proxy.hostname")
             proxy_platform = c.strget("proxy.platform")
@@ -485,11 +526,35 @@ class UIXpraClient(ClientBaseClass):
 
     def _process_startup_complete(self, packet):
         log("all the existing windows and system trays have been received")
-        XpraClientBase._process_startup_complete(self, packet)
+        super()._process_startup_complete(packet)
         gui_ready()
         if self.tray:
             self.tray.ready()
         self.send_info_request()
+        msg = "running"
+        try:
+            windows = tuple(self._id_to_window.values())
+        except AttributeError:
+            pass
+        else:
+            trays = sum(1 for w in windows if w.is_tray())
+            wins = sum(1 for w in windows if not w.is_tray())
+            if wins:
+                msg += ", %i window%s" % (wins, engs(wins))
+            if trays:
+                msg += ", %i tray%s" % (trays, engs(trays))
+        log.info(msg)
+
+    def _process_new_window(self, packet):
+        window = super()._process_new_window(packet)
+        if self.desktop_fullscreen and self._remote_server_mode in ("X11 desktop", "X11 monitor"):
+            from gi.repository import Gdk  # @UnresolvedImport
+            screen = Gdk.Screen.get_default()
+            n = screen.get_n_monitors()
+            monitor = (len(self._id_to_window)-1) % n
+            window.fullscreen_on_monitor(screen, monitor)
+            log("fullscreen_on_monitor: %i", monitor)
+        return window
 
 
     def handshake_complete(self):
@@ -532,31 +597,29 @@ class UIXpraClient(ClientBaseClass):
             "sharing", "sharing-toggle", "lock", "lock-toggle",
             "start-new-commands", "client-shutdown", "webcam",
             "bandwidth-limit", "clipboard-limits",
-            "xdg-menu",
+            "xdg-menu", "monitors",
             ):
             setattr(self, "server_%s" % setting.replace("-", "_"), value)
         else:
             log.info("unknown server setting changed: %s=%s", setting, repr_ellipsized(bytestostr(value)))
             return
         log("_process_setting_change: %s=%s", setting, value)
-        #xdg-menu is too big to log, and we have to update our attribute:
-        if setting=="xdg-menu":
-            self.server_xdg_menu = value
-        else:
+        #thse are too big to log
+        if setting not in ("xdg-menu", "monitors"):
             log.info("server setting changed: %s=%s", setting, repr_ellipsized(value))
         self.server_setting_changed(setting, value)
 
     def server_setting_changed(self, setting, value):
-        log("setting_changed(%s, %s)", setting, value)
+        log("setting_changed(%s, %s)", setting, ellipsizer(value, limit=200))
         cbs = self._on_server_setting_changed.get(setting)
         if cbs:
             for cb in cbs:
-                log("setting_changed(%s, %s) calling %s", setting, value, cb)
+                log("setting_changed(%s, %s) calling %s", setting, ellipsizer(value, limit=200), cb)
                 cb(setting, value)
 
 
     def get_control_commands_caps(self):
-        caps = ["show_session_info", "show_bug_report", "debug"]
+        caps = ["show_session_info", "show_bug_report", "show_menu", "name", "debug"]
         for x in compression.get_enabled_compressors():
             caps.append("enable_"+x)
         for x in packet_encoding.get_enabled_encoders():
@@ -566,12 +629,15 @@ class UIXpraClient(ClientBaseClass):
 
     def _process_control(self, packet):
         command = bytestostr(packet[1])
+        args = packet[2:]
+        log("_process_control(%s)", packet)
         if command=="show_session_info":
-            args = packet[2:]
-            log("calling show_session_info%s on server request", args)
+            log("calling %s%s on server request", self.show_session_info, args)
             self.show_session_info(*args)
         elif command=="show_bug_report":
             self.show_bug_report()
+        elif command=="show_menu":
+            self.show_menu()
         elif command in ("enable_%s" % x for x in compression.get_enabled_compressors()):
             compressor = command.split("_")[1]
             log.info("switching to %s on server request", compressor)
@@ -581,12 +647,11 @@ class UIXpraClient(ClientBaseClass):
             log.info("switching to %s on server request", pe)
             self._protocol.enable_encoder(pe)
         elif command=="name":
-            assert len(args)>=3
-            self.server_session_name = args[2]
+            assert len(args)>=1
+            self.server_session_name = bytestostr(args[0])
             log.info("session name updated from server: %s", self.server_session_name)
             #TODO: reset tray tooltip, session info title, etc..
         elif command=="debug":
-            args = packet[2:]
             if not args:
                 log.warn("not enough arguments for debug control command")
                 return
@@ -676,15 +741,21 @@ class UIXpraClient(ClientBaseClass):
     # keyboard:
     def get_keyboard_caps(self):
         caps = {}
-        if self.readonly or not self.keyboard_helper:
+        kh = self.keyboard_helper
+        if self.readonly or not kh:
             #don't bother sending keyboard info, as it won't be used
             caps["keyboard"] = False
         else:
-            caps.update(self.get_keymap_properties())
+            #legacy, unprefixed:
+            caps.update(kh.get_prefixed_keymap_properties())
+            caps.update({
+                "keyboard"  : True,
+                "keymap"    : kh.get_keymap_properties(),
+                "modifiers" :self.get_current_modifiers(),
+                })
             #show the user a summary of what we have detected:
             self.keyboard_helper.log_keyboard_info()
 
-            caps["modifiers"] = self.get_current_modifiers()
             delay_ms, interval_ms = self.keyboard_helper.key_repeat_delay, self.keyboard_helper.key_repeat_interval
             if delay_ms>0 and interval_ms>0:
                 caps["key_repeat"] = (delay_ms,interval_ms)
@@ -702,19 +773,12 @@ class UIXpraClient(ClientBaseClass):
         if self.keyboard_helper:
             self.keyboard_helper.keymap_changed()
 
-    def get_keymap_properties(self):
-        if not self.keyboard_helper:
-            return {}
-        props = self.keyboard_helper.get_keymap_properties()
-        props["modifiers"] = self.get_current_modifiers()
-        return props
-
     def handle_key_action(self, window, key_event):
         if self.readonly or self.keyboard_helper is None:
-            return
+            return False
         wid = self._window_to_id[window]
-        keylog("handle_key_action(%s, %s) wid=%s", window, key_event, wid)
-        self.keyboard_helper.handle_key_action(window, wid, key_event)
+        keylog(f"handle_key_action({window}, {key_event}) wid={wid}")
+        return self.keyboard_helper.handle_key_action(window, wid, key_event)
 
     def mask_to_names(self, mask):
         if self.keyboard_helper is None:
@@ -729,10 +793,7 @@ class UIXpraClient(ClientBaseClass):
         metadata = typedict(metadata)
         if self.server_is_desktop and self.desktop_fullscreen:
             #force it fullscreen:
-            try:
-                del metadata["size-constraints"]
-            except KeyError:
-                pass
+            metadata.pop("size-constraints", None)
             metadata["fullscreen"] = True
             #FIXME: try to figure out the monitors we go fullscreen on for X11:
             #if POSIX:
@@ -760,7 +821,7 @@ class UIXpraClient(ClientBaseClass):
         #draws spinner on top of the window, or not (plain repaint)
         #depending on whether the server is ok or not
         ok = self.server_ok()
-        log("redraw_spinners() ok=%s", ok)
+        log(f"redraw_spinners() ok={ok}")
         for w in self._id_to_window.values():
             if not w.is_tray():
                 w.spinner(ok)
