@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2016-2021 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2016-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -28,6 +28,31 @@ log = Logger("shadow")
 MULTI_WINDOW = envbool("XPRA_SHADOW_MULTI_WINDOW", True)
 
 
+def parse_geometry(s):
+    try:
+        parts = s.split("@")
+        if len(parts)==1:
+            x = y = 0
+        else:
+            x, y = [int(v.strip(" ")) for v in parts[1].split("x")]
+        w, h = [int(v.strip(" ")) for v in parts[0].split("x")]
+        geometry = [x, y, w, h]
+        screenlog("capture geometry: %s", geometry)
+        return geometry
+    except ValueError:
+        screenlog("failed to parse geometry %r", s, exc_info=True)
+        screenlog.error("Error: invalid display geometry specified: %r", s)
+        screenlog.error(" use the format: WIDTHxHEIGHT@x,y")
+        raise
+
+def parse_geometries(s):
+    g = []
+    for geometry_str in s.split("/"):
+        if geometry_str:
+            g.append(parse_geometry(geometry_str))
+    return g
+
+
 class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
 
     def __init__(self):
@@ -75,7 +100,7 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
     def make_hello(self, source):
         caps = ShadowServerBase.make_hello(self, source)
         caps.update(GTKServerBase.make_hello(self, source))
-        if source.wants_features:
+        if "features" in source.wants:
             caps["screen_sizes"] = get_screen_sizes()
         return caps
 
@@ -85,13 +110,14 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
         if not self.mapped:
             self.refresh_timer = None
             return False
+        self.refresh_window_models()
         if self.capture:
             try:
                 if not self.capture.refresh():
                     #capture doesn't have any screen updates,
                     #so we can skip calling damage
                     #(this shortcut is only used with nvfbc)
-                    return False
+                    return True
             except TransientCodecException as e:
                 log("refresh()", exc_info=True)
                 log.warn("Warning: transient codec exception:")
@@ -104,9 +130,12 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
                 log.warn(" %s", e)
                 self.recreate_window_models()
                 return False
+        self.refresh_windows()
+        return True
+
+    def refresh_windows(self):
         for window in self._id_to_window.values():
             self.refresh_window(window)
-        return True
 
 
     ############################################################################
@@ -169,31 +198,13 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
         match_str = None
         multi_window = MULTI_WINDOW
         geometries = None
-        def parse_geometry(s):
-            try:
-                parts = s.split("@")
-                if len(parts)==1:
-                    x = y = 0
-                else:
-                    x, y = [int(v.strip(" ")) for v in parts[1].split("x")]
-                w, h = [int(v.strip(" ")) for v in parts[0].split("x")]
-                geometry = [x, y, w, h]
-                screenlog("capture geometry: %s", geometry)
-                return geometry
-            except ValueError:
-                screenlog("failed to parse geometry %r", s, exc_info=True)
-                screenlog.error("Error: invalid display geometry specified: %r", s)
-                screenlog.error(" use the format: WIDTHxHEIGHT@x,y")
-                raise
-        def parse_geometries(s):
-            g = []
-            for geometry_str in s.split("/"):
-                if geometry_str:
-                    g.append(parse_geometry(geometry_str))
-            return g
         if "=" in self.display_options:
             #parse the display options as a dictionary:
             opt_dict = parse_simple_dict(self.display_options)
+            windows = opt_dict.get("windows")
+            if windows:
+                self.window_matches = windows.split("/")
+                return self.makeDynamicWindowModels()
             match_str = opt_dict.get("plug")
             multi_window = parse_bool("multi-window", opt_dict.get("multi-window", multi_window))
             geometries_str = opt_dict.get("geometry")
@@ -202,14 +213,11 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
         else:
             try:
                 geometries = parse_geometries(self.display_options)
-            except:
+            except Exception:
                 match_str = self.display_options
         if not multi_window or geometries:
-            for geometry in (geometries or (None,)):
-                model = model_class(self.root, self.capture)
-                model.title = display_name
-                if geometry:
-                    model.geometry = geometry
+            for geometry in (geometries or (self.root.get_geometry()[:4],)):
+                model = model_class(self.root, self.capture, display_name, geometry)
                 models.append(model)
             return models
         found = []
@@ -223,17 +231,57 @@ class GTKShadowServerBase(ShadowServerBase, GTKServerBase):
             if match_str and not(title in match_str or plug_name in match_str):
                 screenlog.info(" skipped monitor %s", plug_name or title)
                 continue
-            model = model_class(self.root, self.capture)
-            model.title = title
-            model.geometry = (x, y, width, height)
+            geometry = (x, y, width, height)
+            model = model_class(self.root, self.capture, title, geometry)
             models.append(model)
-            screenlog("monitor %i: %10s geometry=%s, scale factor=%s", i, title, model.geometry, scale_factor)
+            screenlog("monitor %i: %10s geometry=%s, scale factor=%s", i, title, geometry, scale_factor)
         screenlog("makeRootWindowModels()=%s", models)
         if not models and match_str:
             screenlog.warn("Warning: no monitors found matching %r", match_str)
             screenlog.warn(" only found: %s", csv(found))
         return models
 
+    def makeDynamicWindowModels(self):
+        assert self.window_matches
+        raise NotImplementedError("dynamic window shadow is not implemented on this platform")
+
+    def refresh_window_models(self):
+        if not self.window_matches or not server_features.windows:
+            return
+        #update the window models which may have changed,
+        #some may have disappeared, new ones created,
+        #or they may just have changed their geometry:
+        try:
+            windows = self.makeDynamicWindowModels()
+        except Exception as e:
+            log("refresh_window_models()", exc_info=True)
+            log.error("Error refreshing window models")
+            log.error(" %s", e)
+            return
+        #build a map of window identifier -> window model:
+        xid_to_window = dict((window.get_id(), window) for window in windows)
+        log("xid_to_window(%s)=%s", windows, xid_to_window)
+        sources = self.window_sources()
+        for wid, window in tuple(self._id_to_window.items()):
+            xid = window.get_id()
+            new_model = xid_to_window.pop(xid, None)
+            if new_model is None:
+                #window no longer exists:
+                self._remove_window(window)
+                continue
+            resized = window.geometry[2:]!=new_model.geometry[2:]
+            window.geometry = new_model.geometry
+            if resized:
+                #it has been resized:
+                window.geometry = new_model.geometry
+                window.notify("size-hints")
+                for ss in sources:
+                    ss.resize_window(wid, window, window.geometry[2], window.geometry[3])
+        #any models left are new windows:
+        for window in xid_to_window.values():
+            self._add_new_window(window)
+            self.refresh_window(window)
+
 
     def _adjust_pointer(self, proto, wid, opointer):
         window = self._id_to_window.get(wid)
