@@ -1,24 +1,58 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2012-2019 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2012-2022 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import socket
+from collections import namedtuple
 
-from xpra.util import prettify_plug_name
-from xpra.os_util import get_generic_os_name, load_binary_file
+from xpra.os_util import (
+    get_generic_os_name, do_get_generic_os_name,
+    load_binary_file, get_linux_distribution,
+    )
 from xpra.platform.paths import get_icon_filename
 from xpra.log import Logger
 
 log = Logger("shadow")
 
 
-class RootWindowModel(object):
+def get_os_icons():
+    try:
+        from PIL import Image  # pylint: disable=import-outside-toplevel
+    except ImportError:
+        return ()
+    filename = (get_generic_os_name() or "").lower()+".png"
+    icon_name = get_icon_filename(filename)
+    if not icon_name:
+        log(f"get_os_icons() no icon matching {filename!r}")
+        return ()
+    try:
+        img = Image.open(icon_name)
+        log(f"Image({icon_name})={img}")
+        if img:
+            icon_data = load_binary_file(icon_name)
+            if not icon_data:
+                log(f"icon {icon_name} not found")
+                return ()
+            w, h = img.size
+            img.close()
+            icon = (w, h, "png", icon_data)
+            icons = (icon,)
+            return icons
+    except Exception:   # pragma: no cover
+        log.error(f"Error: failed to load window icon {icon_name!r}", exc_info=True)
+    return ()
 
-    def __init__(self, root_window, capture=None):
+
+class RootWindowModel:
+    __slots__ = ("window", "title", "geometry", "capture",
+                 "property_names", "dynamic_property_names", "internal_property_names",
+                 "signal_listeners")
+    def __init__(self, root_window, capture=None, title="", geometry=None):
         self.window = root_window
-        self.geometry = root_window.get_geometry()[:4]
+        self.title = title
+        self.geometry = geometry
         self.capture = capture
         self.property_names = [
             "title", "class-instance",
@@ -28,11 +62,12 @@ class RootWindowModel(object):
             ]
         self.dynamic_property_names = []
         self.internal_property_names = ["content-type"]
+        self.signal_listeners = {}
 
     def __repr__(self):
-        return "RootWindowModel(%s : %24s)" % (self.capture, self.geometry)
+        return f"RootWindowModel({self.capture} : {str(self.geometry):24})"
 
-    def get_info(self):
+    def get_info(self) -> dict:
         info = {}
         c = self.capture
         if c:
@@ -106,7 +141,7 @@ class RootWindowModel(object):
             return getattr(self, attr_name)
         #otherwise fallback to default behaviour:
         if prop=="title":
-            return prettify_plug_name(self.window.get_screen().get_display().get_name())
+            return self.title
         if prop=="client-machine":
             return socket.gethostname()
         if prop=="window-type":
@@ -129,27 +164,18 @@ class RootWindowModel(object):
                 "base-size" : size,
                 }
         if prop=="class-instance":
-            osn = get_generic_os_name()
-            return ("xpra-%s" % osn, "Xpra-%s" % osn.upper())
+            osn = do_get_generic_os_name()
+            if osn=="Linux":
+                try:
+                    osn += "-"+get_linux_distribution()[0].replace(" ", "-")
+                except Exception:   # pragma: no cover
+                    pass
+            return f"xpra-{osn.lower()}", f"Xpra {osn.replace('-', ' ')}"
         if prop=="icons":
-            try:
-                icon_name = get_icon_filename((get_generic_os_name() or "").lower()+".png")
-                from PIL import Image
-                img = Image.open(icon_name)
-                log("Image(%s)=%s", icon_name, img)
-                if img:
-                    icon_data = load_binary_file(icon_name)
-                    assert icon_data
-                    w, h = img.size
-                    icon = (w, h, "png", icon_data)
-                    icons = (icon,)
-                    return icons
-            except:
-                log("failed to return window icon")
-            return ()
+            return get_os_icons()
         if prop=="content-type":
             return "desktop"
-        raise ValueError("invalid property: %s" % prop)
+        raise ValueError(f"invalid property {prop!r}")
 
     def get(self, name, default_value=None):
         try:
@@ -158,12 +184,28 @@ class RootWindowModel(object):
             log("get(%s, %s) %s on %s", name, default_value, e, self)
             return default_value
 
+    def notify(self, prop):
+        if prop not in self.dynamic_property_names:
+            log.warn(f"Warning: ignoring notify for {prop!r}")
+            return
+        PSpec = namedtuple("PSpec", "name")
+        pspec = PSpec(name=prop)
+        listeners = self.signal_listeners.get(prop)
+        for listener, *args in listeners:
+            try:
+                listener(self, pspec, *args)
+            except Exception:
+                log.error(f"Error on {prop!r} signal listener {listener}", exc_info=True)
 
-    def managed_connect(self, *args):
-        log.warn("ignoring managed signal connect request: %s", args)
+    def managed_connect(self, signal, *args):   # pragma: no cover
+        self.connect(signal, *args)
 
-    def connect(self, *args):
-        log.warn("ignoring signal connect request: %s", args)
+    def connect(self, signal, *args):           # pragma: no cover
+        prop = signal.split(":")[-1]        #notify::geometry
+        if prop not in self.dynamic_property_names:
+            log.warn(f"Warning: ignoring signal connect request: {args}")
+            return
+        self.signal_listeners.setdefault(prop, []).append(args)
 
-    def disconnect(self, *args):
-        log.warn("ignoring signal disconnect request: %s", args)
+    def disconnect(self, *args):        # pragma: no cover
+        log.warn(f"Warning: ignoring signal disconnect request: {args}")
