@@ -1,10 +1,11 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2021 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import sys
+from time import monotonic
 from collections import deque
 from threading import Lock
 from gi.repository import GObject
@@ -12,9 +13,9 @@ from gi.repository import GObject
 from xpra.sound.sound_pipeline import SoundPipeline
 from xpra.gtk_common.gobject_util import one_arg_signal
 from xpra.sound.gstreamer_util import (
-    plugin_str, get_decoder_elements,
+    plugin_str, get_decoder_elements, has_plugins,
     get_queue_time, normv, get_decoders,
-    get_default_sink, get_sink_plugins,
+    get_default_sink_plugin, get_sink_plugins,
     MP3, CODEC_ORDER, gst, QUEUE_LEAK,
     GST_QUEUE_NO_LEAK, MS_TO_NS, DEFAULT_SINK_PLUGIN_OPTIONS,
     GST_FLOW_OK,
@@ -22,7 +23,6 @@ from xpra.sound.gstreamer_util import (
 from xpra.net.compression import decompress_by_name
 from xpra.scripts.config import InitExit
 from xpra.util import csv, envint, envbool
-from xpra.os_util import monotonic_time
 from xpra.make_thread import start_thread
 from xpra.log import Logger
 
@@ -48,6 +48,7 @@ GRACE_PERIOD = envint("XPRA_SOUND_GRACE_PERIOD", 2000)
 MARGIN = max(0, min(200, envint("XPRA_SOUND_MARGIN", 50)))
 #how high we push up the min-level to prevent underruns:
 UNDERRUN_MIN_LEVEL = max(0, envint("XPRA_SOUND_UNDERRUN_MIN_LEVEL", 150))
+CLOCK_SYNC = envbool("XPRA_CLOCK_SYNC", False)
 
 
 GST_FORMAT_BYTES = 2
@@ -68,7 +69,7 @@ class SoundSink(SoundPipeline):
 
     def __init__(self, sink_type=None, sink_options=None, codecs=(), codec_options=None, volume=1.0):
         if not sink_type:
-            sink_type = get_default_sink()
+            sink_type = get_default_sink_plugin()
         if sink_type not in get_sink_plugins():
             raise InitExit(1, "invalid sink: %s" % sink_type)
         matching = [x for x in CODEC_ORDER if (x in codecs and x in get_decoders())]
@@ -87,6 +88,7 @@ class SoundSink(SoundPipeline):
         self.levels = deque(maxlen=100)
         self.volume = None
         self.src    = None
+        self.sink   = None
         self.queue  = None
         self.normal_volume = volume
         self.target_volume = volume
@@ -99,8 +101,8 @@ class SoundSink(SoundPipeline):
         self.last_underrun = 0
         self.last_overrun = 0
         self.refill = True
-        self.last_max_update = monotonic_time()
-        self.last_min_update = monotonic_time()
+        self.last_max_update = monotonic()
+        self.last_min_update = monotonic()
         self.level_lock = Lock()
         pipeline_els = []
         appsrc_el = ["appsrc",
@@ -128,6 +130,11 @@ class SoundSink(SoundPipeline):
                                           "max-size-time=%s" % QUEUE_TIME,
                                           "leaky=%s" % QUEUE_LEAK]))
         pipeline_els.append("volume name=volume volume=0")
+        if CLOCK_SYNC:
+            if not has_plugins("clocksync"):
+                log.warn("Warning: cannot enable clocksync, element not found")
+            else:
+                pipeline_els.append("clocksync")
         sink_attributes = SINK_SHARED_DEFAULT_ATTRIBUTES.copy()
         #anything older than this may cause problems (ie: centos 6.x)
         #because the attributes may not exist
@@ -139,12 +146,14 @@ class SoundSink(SoundPipeline):
             sink_attributes.update(v)
         if sink_options:
             sink_attributes.update(sink_options)
+        sink_attributes["name"] = "sink"
         sink_str = plugin_str(sink_type, sink_attributes)
         pipeline_els.append(sink_str)
         if not self.setup_pipeline_and_bus(pipeline_els):
             return
         self.volume = self.pipeline.get_by_name("volume")
         self.src    = self.pipeline.get_by_name("src")
+        self.sink   = self.pipeline.get_by_name("sink")
         self.queue  = self.pipeline.get_by_name("queue")
         if self.queue:
             if QUEUE_SILENT:
@@ -154,18 +163,19 @@ class SoundSink(SoundPipeline):
                 self.queue.connect("underrun", self.queue_underrun)
                 self.queue.connect("running", self.queue_running)
                 self.queue.connect("pushing", self.queue_pushing)
+        self.init_file(codec)
 
-    def __repr__(self):
+    def __repr__(self):  #pylint: disable=arguments-differ
         return "SoundSink('%s' - %s)" % (self.pipeline_str, self.state)
 
     def cleanup(self):
-        SoundPipeline.cleanup(self)
+        super().cleanup()
         self.cancel_volume_timer()
         self.sink_type = ""
         self.src = None
 
     def start(self):
-        SoundPipeline.start(self)
+        super().start()
         self.timeout_add(UNMUTE_DELAY, self.start_adjust_volume)
 
 
@@ -210,7 +220,7 @@ class SoundSink(SoundPipeline):
         return True
 
     def queue_underrun(self, *_args):
-        now = monotonic_time()
+        now = monotonic()
         if self.queue_state=="starting" or 1000*(now-self.start_time)<GRACE_PERIOD:
             gstlog("ignoring underrun during startup")
             return True
@@ -231,23 +241,23 @@ class SoundSink(SoundPipeline):
         return True
 
     def get_level_range(self, mintime=2, maxtime=10):
-        now = monotonic_time()
+        now = monotonic()
         filtered = [v for t,v in tuple(self.levels) if (now-t)>=mintime and (now-t)<=maxtime]
         if len(filtered)>=10:
             maxl = max(filtered)
             minl = min(filtered)
             #range of the levels recorded:
             return maxl-minl
-        return GST_FLOW_OK
+        return 0
 
     def queue_overrun(self, *_args):
-        now = monotonic_time()
+        now = monotonic()
         if self.queue_state=="starting" or 1000*(now-self.start_time)<GRACE_PERIOD:
             gstlog("ignoring overrun during startup")
             return True
         clt = self.queue.get_property("current-level-time")//MS_TO_NS
         log("queue_overrun level=%ims", clt)
-        now = monotonic_time()
+        now = monotonic()
         #grace period of recording overruns:
         #(because when we record an overrun, we lower the max-time,
         # which causes more overruns!)
@@ -261,7 +271,7 @@ class SoundSink(SoundPipeline):
     def set_min_level(self):
         if not self.queue:
             return
-        now = monotonic_time()
+        now = monotonic()
         elapsed = now-self.last_min_update
         lrange = self.get_level_range()
         log("set_min_level() lrange=%i, elapsed=%i", lrange, elapsed)
@@ -294,7 +304,7 @@ class SoundSink(SoundPipeline):
     def set_max_level(self):
         if not self.queue:
             return
-        now = monotonic_time()
+        now = monotonic()
         elapsed = now-self.last_max_update
         if elapsed<1:
             #not more than once a second
@@ -342,7 +352,7 @@ class SoundSink(SoundPipeline):
         return GST_FLOW_OK
 
     def get_info(self) -> dict:
-        info = SoundPipeline.get_info(self)
+        info = super().get_info()
         if QUEUE_TIME>0 and self.queue:
             clt = self.queue.get_property("current-level-time")
             qmax = self.queue.get_property("max-size-time")
@@ -356,6 +366,17 @@ class SoundSink(SoundPipeline):
                              "underruns"    : self.underruns,
                              "state"        : self.queue_state,
                              }
+        info["sink"] = self.get_element_properties(
+            self.sink,
+            "buffer-time", "latency-time",
+            #"next_sample", "eos_rendering",
+            "async", "blocksize",
+            "enable-last-sample",
+            "max-bitrate", "max-lateness",
+            #"processing-deadline",
+            "qos", "render-delay", "sync",
+            "throttle-time", "ts-offset",
+            )
         return info
 
     def can_push_buffer(self):
@@ -374,7 +395,7 @@ class SoundSink(SoundPipeline):
         compress = metadata.get("compress")
         if not compress:
             return data
-        assert compress in ("lz4", "lzo")
+        assert compress in ("lz4", )
         v = decompress_by_name(data, compress)
         #log("decompressed %s data: %i bytes into %i bytes", compress, len(data), len(v))
         return v
@@ -402,6 +423,7 @@ class SoundSink(SoundPipeline):
     def do_add_data(self, data, metadata=None):
         #having a timestamp causes problems with the queue and overruns:
         log("do_add_data(%s bytes, %s) queue_state=%s", len(data), metadata, self.queue_state)
+        self.save_to_file(data)
         buf = gst.Buffer.new_allocate(None, len(data), None)
         buf.fill(0, data)
         if metadata:
@@ -427,7 +449,7 @@ class SoundSink(SoundPipeline):
             return
         clt = q.get_property("current-level-time")//MS_TO_NS
         log("pushed %5i bytes, new buffer level: %3ims, queue state=%s", len(data), clt, self.queue_state)
-        now = monotonic_time()
+        now = monotonic()
         self.levels.append((now, clt))
 
     def push_buffer(self, buf):
@@ -476,7 +498,7 @@ def main():
             parts = filename.split(".")
             if len(parts)>1:
                 extension = parts[-1]
-                if extension.lower() in codecs:
+                if extension.lower() in decoders:
                     codec = extension.lower()
                     print("guessed codec %s from file extension %s" % (codec, extension))
             if codec is None:
