@@ -20,18 +20,22 @@ from xpra.x11.gtk_x11.gdk_bindings import (
 from xpra.gtk_common.error import XError
 from xpra.clipboard.clipboard_core import (
     ClipboardProxyCore, TEXT_TARGETS,
-    must_discard, must_discard_extra, _filter_targets,
+    must_discard, must_discard_extra,
     )
 from xpra.clipboard.clipboard_timeout_helper import ClipboardTimeoutHelper, CONVERT_TIMEOUT
 from xpra.x11.bindings.window_bindings import ( #@UnresolvedImport
     constants, PropertyError,                   #@UnresolvedImport
     X11WindowBindings,                          #@UnresolvedImport
     )
-from xpra.os_util import bytestostr, strtobytes
+from xpra.x11.bindings.res_bindings import ResBindings #@UnresolvedImport
+from xpra.os_util import bytestostr, strtobytes, memoryview_to_bytes
 from xpra.util import csv, repr_ellipsized, ellipsizer, first_time
 from xpra.log import Logger
 
 X11Window = X11WindowBindings()
+XRes = ResBindings()
+if not XRes.check_xres():
+    XRes = None
 
 log = Logger("x11", "clipboard")
 
@@ -294,6 +298,7 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
             with xsync:
                 owner = X11Window.XGetSelectionOwner(self._selection)
                 if owner==self.xid:
+                    self.owned = True
                     log("claim() we already own the '%s' selection", self._selection)
                     return
                 setsel = X11Window.XSetSelectionOwner(self.xid, self._selection, time)
@@ -301,7 +306,6 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
                 self.owned = owner==self.xid
                 log("claim_selection: set selection owner returned %s, owner=%#x, owned=%s",
                     setsel, owner, self.owned)
-                event_mask = StructureNotifyMask
                 if not self.owned:
                     log.warn("Warning: we failed to get ownership of the '%s' clipboard selection", self._selection)
                     return
@@ -309,9 +313,10 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
                 log("claim_selection: sending message to root window")
                 root = get_default_root_window()
                 root_xid = root.get_xid()
+                event_mask = StructureNotifyMask
                 X11Window.sendClientMessage(root_xid, root_xid, False, event_mask, "MANAGER",
                                   time or CurrentTime, self._selection, self.xid)
-                log("claim_selection: done, owned=%s", self.owned)
+                log(f"claim_selection {self._selection} done")
         except Exception:
             log("failed to claim selection '%s'", self._selection, exc_info=True)
             raise
@@ -335,17 +340,25 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
         return None
 
     def get_wininfo(self, xid):
+        wininfo = [f"xid={xid:x}"]
+        if XRes:
+            with xswallow:
+                pid = XRes.get_pid(xid)
+                if pid:
+                    wininfo.append(f"pid={pid}")
         with xswallow:
             title = self.get_wintitle(xid)
             if title:
-                return "'%s'" % title
+                wininfo.insert(0, title)
+                return wininfo
         with xswallow:
             while xid:
                 title = self.get_wintitle(xid)
                 if title:
-                    return "child of '%s'" % title
+                    wininfo.append(f"child of {title!r}")
+                    return wininfo
                 xid = X11Window.getParent(xid)
-        return hex(xid)
+        return wininfo
 
     ############################################################################
     # forward local requests to the remote clipboard:
@@ -360,8 +373,8 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
         wininfo = self.get_wininfo(requestor.get_xid())
         prop = event.property
         target = str(event.target)
-        log("clipboard request for %s from window %#x: %s, target=%s, prop=%s",
-            self._selection, requestor.get_xid(), wininfo, target, prop)
+        log("clipboard request for %s from window %s, target=%s, prop=%s",
+            self._selection, wininfo, target, prop)
         if not target:
             log.warn("Warning: ignoring clipboard request without a TARGET")
             log.warn(" coming from %s", wininfo)
@@ -375,9 +388,10 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
         if not self._enabled:
             nodata()
             return
-        if wininfo and wininfo.strip("'") in BLACKLISTED_CLIPBOARD_CLIENTS:
-            if first_time("clipboard-blacklisted:%s" % wininfo.strip("'")):
-                log.warn("receiving clipboard requests from blacklisted client %s", wininfo)
+        blacklisted = tuple(client for client in BLACKLISTED_CLIPBOARD_CLIENTS if client in wininfo)
+        if blacklisted:
+            if first_time(f"clipboard-blacklisted:{blacklisted}"):
+                log.warn(f"receiving clipboard requests from blacklisted client {csv(wininfo)}")
                 log.warn(" all requests will be silently ignored")
             log("responding with nodata for blacklisted client '%s'", wininfo)
             return
@@ -462,14 +476,7 @@ class ClipboardProxy(ClipboardProxyCore, GObject.GObject):
                 return
             with xsync:
                 if data is not None:
-                    if isinstance(data, str):
-                        #the data is already in the correct format,
-                        #but the cython bindings require real 'bytes'
-                        try:
-                            data = strtobytes(data)
-                        except UnicodeEncodeError:
-                            data = data.encode("utf8")
-                    X11Window.XChangeProperty(xid, prop, dtype, dformat, data)
+                    X11Window.XChangeProperty(xid, prop, dtype, dformat, memoryview_to_bytes(data))
                 else:
                     #maybe even delete the property?
                     #X11Window.XDeleteProperty(xid, prop)
